from ovos_workshop.app import OVOSAbstractApplication
from ovos_workshop.decorators import *
from ovos_workshop.decorators.killable import killable_event, \
    AbortEvent, AbortQuestion
from ovos_workshop.decorators.layers import IntentLayers



# Copyright 2017 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
from os.path import join, expanduser, isdir
from typing import TextIO

from ovos_config.locations import get_xdg_data_save_path
from ovos_config.meta import get_xdg_base
from ovos_utils.log import log_deprecation


class FileSystemAccess:
    def __init__(self, path: str):
        """
        Create a filesystem in a valid location.
        @param path: path basename/name of the module requesting a filesystem
        """
        self.path = self.__init_path(path)

    @staticmethod
    def __init_path(path: str):
        """
        Initialize a directory for filesystem access.
        @param path: path basename to initialize
        @return: validated existing path for this filesystem
        """
        if not isinstance(path, str) or len(path) == 0:
            raise ValueError("path must be initialized as a non empty string")

        old_path = expanduser(f'~/.{get_xdg_base()}/{path}')
        xdg_path = expanduser(f'{get_xdg_data_save_path()}/filesystem/{path}')
        # Migrate from the old location if it still exists
        if isdir(old_path) and not isdir(xdg_path):
            log_deprecation(f"Settings at {old_path} will be ignored", "0.1.0")
            shutil.move(old_path, xdg_path)

        if not isdir(xdg_path):
            os.makedirs(xdg_path)
        return xdg_path

    def open(self, filename: str, mode: str) -> TextIO:
        """
        Open the requested file in this FileSystem in the requested mode.
        @param filename: string filename, relative to this FileSystemAccess
        @param mode: mode to open file with (i.e. `rb`, `w+`)
        @return: TextIO object for the requested file in the requested mode
        """
        file_path = join(self.path, filename)
        return open(file_path, mode)

    def exists(self, filename: str) -> bool:
        """
        Check if a file exists in the namespace.
        @param filename: string filename, relative to this FileSystemAccess
        @return: True if the filename exists, else False
        """
        return os.path.exists(join(self.path, filename))



# backwards compat imports
from ovos_utils.ocp import MediaType, PlayerState, MediaState, MatchConfidence,\
    PlaybackType, PlaybackMode, LoopState, TrackState


def ocp_search():
    """
    Decorator for adding a method as a common play search handler.
    Decorated methods should either yield or return a list of dict results:
    {
      "media_type": <MediaType>,
      "playback": <PlaybackType>,
      "image": <(optional) str image/cover art URI>,
      "skill_icon": <(optional) str skill icon URI>,
      "bg_image": <(optional) str background image URI>,
      "uri": <str media URI>,
      "title": <str media title>,
      "artist": <str media artist/author>,
      "length": <(optional) int media length in milliseconds>,
      "match_confidence": <int 0-100 confidence this result matches request>
    }
    """

    def real_decorator(func):
        # Store the flag inside the function
        # This will be used later to identify the method
        if not hasattr(func, 'is_ocp_search_handler'):
            func.is_ocp_search_handler = True

        return func

    return real_decorator


def ocp_play():
    """
    Decorator for adding a method to handle media playback.
    """

    def real_decorator(func):
        # Store the flag inside the function
        # This will be used later to identify the method
        if not hasattr(func, 'is_ocp_playback_handler'):
            func.is_ocp_playback_handler = True

        return func

    return real_decorator


def ocp_previous():
    """
    Decorator for adding a method to handle requests to skip backward.
    """

    def real_decorator(func):
        # Store the flag inside the function
        # This will be used later to identify the method
        if not hasattr(func, 'is_ocp_prev_handler'):
            func.is_ocp_prev_handler = True

        return func

    return real_decorator


def ocp_next():
    """
    Decorator for adding a method to handle requests to skip forward.
    """

    def real_decorator(func):
        # Store the flag inside the function
        # This will be used later to identify the method
        if not hasattr(func, 'is_ocp_next_handler'):
            func.is_ocp_next_handler = True

        return func

    return real_decorator


def ocp_pause():
    """
    Decorator for adding a method to handle requests to pause playback.
    """

    def real_decorator(func):
        # Store the flag inside the function
        # This will be used later to identify the method
        if not hasattr(func, 'is_ocp_pause_handler'):
            func.is_ocp_pause_handler = True

        return func

    return real_decorator


def ocp_resume():
    """
    Decorator for adding a method to handle requests to resume playback.
    """

    def real_decorator(func):
        # Store the flag inside the function
        # This will be used later to identify the method
        if not hasattr(func, 'is_ocp_resume_handler'):
            func.is_ocp_resume_handler = True

        return func

    return real_decorator


def ocp_featured_media():
    """
    Decorator for adding a method to handle requests to provide featured media.
    """

    def real_decorator(func):
        # Store the flag inside the function
        # This will be used later to identify the method
        if not hasattr(func, 'is_ocp_featured_handler'):
            func.is_ocp_featured_handler = True

        return func

    return real_decorator




import os
from inspect import signature
from threading import Event
from typing import List

from ovos_utils import camel_case_split
from ovos_utils.log import LOG

from ovos_bus_client import Message
from ovos_classifiers.skovos.features import KeywordFeatures
from ovos_config.locations import get_xdg_cache_save_path
from ovos_workshop.skills.ovos import OVOSSkill

# backwards compat imports, do not delete, skills import from here
from ovos_workshop.decorators.ocp import ocp_play, ocp_next, ocp_pause, ocp_resume, ocp_search, \
    ocp_previous, ocp_featured_media
from ovos_utils.ocp import MediaType, MediaState, MatchConfidence, \
    PlaybackType, PlaybackMode, PlayerState, LoopState, TrackState


def get_non_properties(obj):
    """Get attibutes that are not properties from object.

    Will return members of object class along with bases down to MycroftSkill.

    Args:
        obj: object to scan

    Returns:
        Set of attributes that are not a property.
    """

    def check_class(cls):
        """Find all non-properties in a class."""
        # Current class
        d = cls.__dict__
        np = [k for k in d if not isinstance(d[k], property)]
        # Recurse through base classes excluding MycroftSkill and object
        for b in [b for b in cls.__bases__ if b not in (object, OVOSSkill)]:
            np += check_class(b)
        return np

    return set(check_class(obj.__class__))


class OVOSCommonPlaybackSkill(OVOSSkill):
    """ To integrate with the OpenVoiceOS Common Playback framework
    skills should use this base class and the companion decorators

    @ocp_search
    def ...

    @ocp_play
    def ...

    The class makes the skill available to queries from OCP and no special
    vocab for starting playback is needed.
    """

    def __init__(self, supported_media=None, skill_icon="", *args, **kwargs):
        self.supported_media = supported_media or [MediaType.GENERIC]
        skill_name = camel_case_split(self.__class__.__name__)
        alt = skill_name.replace(" skill", "").replace(" Skill", "")
        self.skill_aliases = [skill_name, alt]

        self._search_handlers = []  # added via decorators
        self._featured_handlers = []  # added via decorators
        self._current_query = None
        self.__playback_handler = None
        self.__pause_handler = None
        self.__next_handler = None
        self.__prev_handler = None
        self.__resume_handler = None
        self._stop_event = Event()
        self._playing = Event()
        # TODO new default icon
        self.skill_icon = skill_icon or ""

        self.ocp_matchers = {}
        super().__init__(*args, **kwargs)

    @property
    def ocp_cache_dir(self):
        """path to cached .csv file with ocp entities data
        this file needs to be available in ovos-core

        NB: ovos-docker needs a shared volume
        """
        os.makedirs(f"{get_xdg_cache_save_path()}/OCP", exist_ok=True)
        return f"{get_xdg_cache_save_path()}/OCP"

    def bind(self, bus):
        """Overrides the normal bind method.

        Adds handlers for play:query and play:start messages allowing
        interaction with the playback control skill.

        This is called automatically during setup, and
        need not otherwise be used.
        """
        if bus:
            super().bind(bus)
            self.add_event('ovos.common_play.query',
                           self.__handle_ocp_query)
            self.add_event(f'ovos.common_play.query.{self.skill_id}',
                           self.__handle_ocp_query)
            self.add_event('ovos.common_play.featured_tracks.play',
                           self.__handle_ocp_featured)
            self.add_event('ovos.common_play.skills.get',
                           self.__handle_ocp_skills_get)
            self.add_event(f'ovos.common_play.{self.skill_id}.play',
                           self.__handle_ocp_play)
            self.add_event(f'ovos.common_play.{self.skill_id}.pause',
                           self.__handle_ocp_pause)
            self.add_event(f'ovos.common_play.{self.skill_id}.resume',
                           self.__handle_ocp_resume)
            self.add_event(f'ovos.common_play.{self.skill_id}.next',
                           self.__handle_ocp_next)
            self.add_event(f'ovos.common_play.{self.skill_id}.previous',
                           self.__handle_ocp_prev)
            self.add_event(f'ovos.common_play.{self.skill_id}.stop',
                           self.__handle_ocp_stop)
            self.add_event("ovos.common_play.search.stop",
                           self.__handle_stop_search)
            self.add_event("mycroft.stop",
                           self.__handle_stop_search)

    def register_media_type(self, media_type: MediaType):
        """ helper instead of editing self.supported_media directly
        will auto-sync changes via bus"""
        if media_type not in self.supported_media:
            self.supported_media.append(media_type)
            LOG.info(f"{self.skill_id} registered type {media_type}")
            self.__handle_ocp_skills_get()

    def __handle_ocp_skills_get(self, message=None):
        """ report skill OCP info

        thumbnail and featured tracks inform the OCP homescreen

        media_type and skill_name help the classifier disambiguate between media_types
            eg, if OCP finds the name of a movie skill in user utterance
                it will search netflix instead of spotify
         """
        message = message or Message("")
        self.bus.emit(
            message.reply('ovos.common_play.announce',
                          {"skill_id": self.skill_id,
                           "skill_name": self.skill_aliases[0],
                           "aliases": self.skill_aliases,
                           "thumbnail": self.skill_icon,
                           "media_type": self.supported_media,
                           "featured_tracks": len(self._featured_handlers) >= 1}))

    def ocp_voc_match(self, utterance, lang=None):
        """uses Aho–Corasick algorithm to match OCP keywords
        this efficiently matches many keywords against an utterance

        OCP keywords are registered via self.register_ocp_keyword

        example usages
            print(self.ocp_voc_match("play metallica"))
            # {'album_name': 'Metallica', 'artist_name': 'Metallica'}

            print(self.ocp_voc_match("play the beatles"))
            # {'album_name': 'The Beatles', 'series_name': 'The Beatles',
            # 'artist_name': 'The Beatles', 'movie_name': 'The Beatles'}

            print(self.ocp_voc_match("play rob zombie"))
            # {'artist_name': 'Rob Zombie', 'album_name': 'Zombie',
            # 'book_name': 'Zombie', 'game_name': 'Zombie', 'movie_name': 'Zombie'}

            print(self.ocp_voc_match("play horror movie"))
            # {'film_genre': 'Horror', 'cartoon_genre': 'Horror', 'anime_genre': 'Horror',
            # 'radio_drama_genre': 'horror', 'video_genre': 'horror',
            # 'book_genre': 'Horror', 'movie_name': 'Horror Movie'}

            print(self.ocp_voc_match("play science fiction"))
            #  {'film_genre': 'Science Fiction', 'cartoon_genre': 'Science Fiction',
            #  'podcast_genre': 'Fiction', 'anime_genre': 'Science Fiction',
            #  'documentary_genre': 'Science', 'book_genre': 'Science Fiction',
            #  'artist_name': 'Fiction', 'tv_channel': 'Science',
            #  'album_name': 'Science Fiction', 'short_film_name': 'Science',
            #  'book_name': 'Science Fiction', 'movie_name': 'Science Fiction'}
        """
        lang = lang or self.lang
        if lang not in self.ocp_matchers:
            return {}
        matches = {}
        for k, v in self.ocp_matchers[lang].match(utterance):
            if k not in matches or len(v) > len(matches[k]):
                matches[k] = v
        return matches

    def load_ocp_keyword_from_csv(self, csv_path: str, lang: str = None):
        """ load entities from a .csv file for usage with self.ocp_voc_match
        see the ocp_entities.csv datatsets for example files built from wikidata SPARQL queries

        examples contents of csv file

            label,entity
            film_genre,swashbuckler film
            film_genre,neo-noir
            film_genre,actual play film
            film_genre,alternate history film
            film_genre,spy film
            ...
        """
        if lang is None:
            for lang in self.native_langs:
                if lang not in self.ocp_matchers:
                    self.ocp_matchers[lang] = KeywordFeatures()
                self.ocp_matchers[lang].load_entities(csv_path)
        else:
            if lang not in self.ocp_matchers:
                self.ocp_matchers[lang] = KeywordFeatures()
            self.ocp_matchers[lang].load_entities(csv_path)

    def export_ocp_keywords_csv(self, csv_path: str = None, lang: str = None,
                                label: str = None):
        """ export entities to a .csv file """
        lang = lang or self.lang
        if lang not in self.ocp_matchers:
            raise RuntimeError(f"no entities registered for lang: {lang}")

        csv_path = csv_path or f"{self.ocp_cache_dir}/{self.skill_id}_{lang}.csv"
        with open(csv_path, "w") as f:
            f.write("label,sample")
            for ent, samples in self.ocp_matchers[lang].entities.items():
                if label is not None and label != ent:
                    continue
                for s in set(samples):
                    f.write(f"\n{ent},{s}")
        LOG.info(f"{self.skill_id} OCP {lang} entities exported to {csv_path}")
        return csv_path

    def register_ocp_keyword(self, media_type: MediaType, label: str,
                             samples: List, langs: List[str] = None):
        """ register strings as native OCP keywords (eg, movie_name, artist_name ...)

        ocp keywords can be efficiently matched with self.ocp_match helper method
        that uses Aho–Corasick algorithm
        """
        samples = list(set(samples))
        langs = langs or self.native_langs
        for l in langs:
            if l not in self.ocp_matchers:
                self.ocp_matchers[l] = KeywordFeatures()
            self.ocp_matchers[l].register_entity(label, samples)

        #  if the label is a valid OCP entity known by the classifier it will help
        #  the classifier disambiguate between media_types
        #  eg, if OCP finds a movie name in user utterances it will
        #      prefer to search netflix instead of spotify

        # NB: we send a file path, bus messages with thousands of entities dont work well
        if len(samples) >= 20:
            csv = f"{self.ocp_cache_dir}/{self.skill_id}_{label}.csv"
            self.export_ocp_keywords_csv(csv, label=label)
            self.bus.emit(
                Message('ovos.common_play.register_keyword',
                        {"skill_id": self.skill_id,
                         "label": label,  # if in OCP_ENTITIES it influences classifier
                         "csv": csv,
                         "media_type": media_type}))
        else:
            self.bus.emit(
                Message('ovos.common_play.register_keyword',
                        {"skill_id": self.skill_id,
                         "label": label,  # if in OCP_ENTITIES it influences classifier
                         "samples": samples,
                         "media_type": media_type}))

    def deregister_ocp_keyword(self, media_type: MediaType, label: str,
                               langs: List[str] = None):
        langs = langs or self.native_langs
        for l in langs:
            if l in self.ocp_matchers:
                self.ocp_matchers[l].deregister_entity(label)

        self.bus.emit(
            Message('ovos.common_play.deregister_keyword',
                    {"skill_id": self.skill_id,
                     "label": label,
                     "media_type": media_type}))

    def _register_decorated(self):
        # register search handlers
        for attr_name in get_non_properties(self):
            method = getattr(self, attr_name)
            if hasattr(method, 'is_ocp_search_handler'):
                if method.is_ocp_search_handler:
                    # TODO this wont accept methods with killable_event
                    #  decorators
                    self._search_handlers.append(method)
            if hasattr(method, 'is_ocp_featured_handler'):
                if method.is_ocp_featured_handler:
                    # TODO this wont accept methods with killable_event
                    #  decorators
                    self._featured_handlers.append(method)
            if hasattr(method, 'is_ocp_playback_handler'):
                if method.is_ocp_playback_handler:
                    # TODO how to handle multiple ??
                    if self.__playback_handler:
                        LOG.warning("multiple declarations of playback "
                                    "handler, replacing previous handler")
                    self.__playback_handler = method
            if hasattr(method, 'is_ocp_pause_handler'):
                if method.is_ocp_pause_handler:
                    # TODO how to handle multiple ??
                    if self.__pause_handler:
                        LOG.warning("multiple declarations of pause "
                                    "handler, replacing previous handler")
                    self.__pause_handler = method
            if hasattr(method, 'is_ocp_next_handler'):
                if method.is_ocp_next_handler:
                    # TODO how to handle multiple ??
                    if self.__next_handler:
                        LOG.warning("multiple declarations of play next "
                                    "handler, replacing previous handler")
                    self.__next_handler = method
            if hasattr(method, 'is_ocp_prev_handler'):
                if method.is_ocp_prev_handler:
                    # TODO how to handle multiple ??
                    if self.__prev_handler:
                        LOG.warning("multiple declarations of play previous "
                                    "handler, replacing previous handler")
                    self.__prev_handler = method
            if hasattr(method, 'is_ocp_resume_handler'):
                if method.is_ocp_resume_handler:
                    # TODO how to handle multiple ??
                    if self.__resume_handler:
                        LOG.warning("multiple declarations of resume playback"
                                    "handler, replacing previous handler")
                    self.__resume_handler = method

        super()._register_decorated()

        # volunteer info to OCP
        self.bus.emit(
            Message('ovos.common_play.announce',
                    {"skill_id": self.skill_id,
                     "skill_name": self.skill_aliases[0],
                     "thumbnail": self.skill_icon,
                     "media_types": self.supported_media,
                     "featured_tracks": len(self._featured_handlers) >= 1}))

    def extend_timeout(self, timeout=0.5):
        """ request more time for searching, limits are defined by
        better-common-play framework, by default max total time is 5 seconds
        per query """
        if self._current_query:
            self.bus.emit(Message("ovos.common_play.query.response",
                                  {"phrase": self._current_query,
                                   "skill_id": self.skill_id,
                                   "skill_name": self.skill_aliases[0],
                                   "thumbnail": self.skill_icon,
                                   "timeout": timeout,
                                   "searching": True}))

    def play_media(self, media, disambiguation=None, playlist=None):
        disambiguation = disambiguation or [media]
        playlist = playlist or [media]
        self.bus.emit(Message("ovos.common_play.play",
                              {"media": media,
                               "disambiguation": disambiguation,
                               "playlist": playlist}))

    # @killable_event("ovos.common_play.stop", react_to_stop=True)
    def __handle_ocp_play(self, message):
        if self.__playback_handler:
            self.__playback_handler(message)
            self.bus.emit(Message("ovos.common_play.player.state",
                                  {"state": PlayerState.PLAYING}))
            self._playing.set()
        else:
            LOG.error(f"Playback requested but {self.skill_id} handler not "
                      "implemented")

    def __handle_ocp_pause(self, message):
        if self.__pause_handler:
            if self.__pause_handler(message):
                self.bus.emit(Message("ovos.common_play.player.state",
                                      {"state": PlayerState.PAUSED}))
        else:
            LOG.error(f"Pause requested but {self.skill_id} handler not "
                      "implemented")

    def __handle_ocp_resume(self, message):
        if self.__resume_handler:
            if self.__resume_handler(message):
                self.bus.emit(Message("ovos.common_play.player.state",
                                      {"state": PlayerState.PLAYING}))
        else:
            LOG.error(f"Resume requested but {self.skill_id} handler not "
                      "implemented")

    def __handle_ocp_next(self, message):
        if self.__next_handler:
            self.__next_handler(message)
        else:
            LOG.error(f"Play Next requested but {self.skill_id} handler not "
                      "implemented")

    def __handle_ocp_prev(self, message):
        if self.__prev_handler:
            self.__prev_handler(message)
        else:
            LOG.error(f"Play Next requested but {self.skill_id} handler not "
                      "implemented")

    def __handle_ocp_stop(self, message):
        # for skills managing their own playback
        if self._playing.is_set():
            self.stop()
            self.gui.release()
            self.bus.emit(Message("ovos.common_play.player.state",
                                  {"state": PlayerState.STOPPED}))
            self._playing.clear()

    def __handle_stop_search(self, message):
        self._stop_event.set()

    # @killable_event("ovos.common_play.search.stop", react_to_stop=True)
    def __handle_ocp_query(self, message: Message):
        """Query skill if it can start playback from given phrase."""
        self._stop_event.clear()
        search_phrase = message.data["phrase"]
        self._current_query = search_phrase
        media_type = message.data.get("question_type",
                                      MediaType.GENERIC)

        if message.msg_type == f'ovos.common_play.query.{self.skill_id}':
            # make message.response work as usual
            message.msg_type = f'ovos.common_play.query'
        elif media_type not in self.supported_media:
            # skip this skill, it doesn't handle this media type
            return

        self.bus.emit(message.reply("ovos.common_play.skill.search_start",
                                    {"skill_id": self.skill_id,
                                     "skill_name": self.skill_aliases[0],
                                     "thumbnail": self.skill_icon, }))

        # invoke the media search handlesr to let the skill perform its search
        found = False
        for handler in self._search_handlers:
            if self._stop_event.is_set():
                break
            # @ocp_search
            # def handle_search(...):
            if len(signature(handler).parameters) == 1:
                # no optional media_type argument
                results = handler(search_phrase) or []
            else:
                results = handler(search_phrase, media_type) or []

            # handler might return a generator or a list
            if isinstance(results, list):
                # inject skill id in individual results, will be needed later
                # for proper playback handling
                for idx, r in enumerate(results):
                    results[idx]["skill_id"] = self.skill_id
                self.bus.emit(message.response({"phrase": search_phrase,
                                                "skill_id": self.skill_id,
                                                "skill_name": self.skill_aliases[0],
                                                "thumbnail": self.skill_icon,
                                                "results": results,
                                                "searching": False}))
                found = True
            else:  # generator, keeps returning results
                for r in results:
                    # inject skill id in individual results, will be needed later
                    # for proper playback handling
                    r["skill_id"] = self.skill_id
                    self.bus.emit(message.response({"phrase": search_phrase,
                                                    "skill_id": self.skill_id,
                                                    "skill_name": self.skill_aliases[0],
                                                    "thumbnail": self.skill_icon,
                                                    "results": [r],
                                                    "searching": False}))
                    found = True
                    if self._stop_event.is_set():
                        break

        if not found:
            # Signal we are done (can't handle it)
            self.bus.emit(message.response({"phrase": search_phrase,
                                            "skill_id": self.skill_id,
                                            "skill_name": self.skill_aliases[0],
                                            "thumbnail": self.skill_icon,
                                            "searching": False}))
        self.bus.emit(message.reply("ovos.common_play.skill.search_end",
                                    {"skill_id": self.skill_id}))

    def __handle_ocp_featured(self, message):
        skill_id = message.data["skill_id"]
        if skill_id != self.skill_id:
            return

        results = []
        for handler in self._featured_handlers:
            try:
                results += list(handler())  # handler might return a generator or a list
            except Exception as e:
                LOG.error(e)

        if not results:
            self.speak_dialog("no.media.available")
        else:
            # inject skill id in individual results
            for idx, r in enumerate(results):
                results[idx]["skill_id"] = self.skill_id
            self.bus.emit(Message("ovos.common_play.skill.play",
                                  {"skill_id": self.skill_id,
                                   "skill_name": self.skill_aliases[0],
                                   "thumbnail": self.skill_icon,
                                   "playlist": results}))

    def default_shutdown(self):
        self.bus.emit(
            Message('ovos.common_play.skills.detach',
                    {"skill_id": self.skill_id}))
        super().default_shutdown()



# DEPRECATED - merged into OVOSSkill, imports for compat onlu
from ovos_workshop.skills.ovos import OVOSSkill, simple_trace, is_classic_core, SkillGUI
from ovos_utils.log import log_deprecation
from ovos_utils.process_utils import RuntimeRequirements


# backwards compat alias
class SkillNetworkRequirements(RuntimeRequirements):
    def __init__(self, *args, **kwargs):
        log_deprecation("Replace with "
                        "`ovos_utils.process_utils.RuntimeRequirements`",
                        "0.1.0")
        super().__init__(*args, **kwargs)


BaseSkill = OVOSSkill  # backwards compat




from abc import ABC

from ovos_config import Configuration

from ovos_bus_client import Message
from ovos_utils.events import get_handler_name
from ovos_utils.log import LOG
from ovos_workshop.resource_files import SkillResources
from ovos_workshop.skills.common_query_skill import CommonQuerySkill
from ovos_workshop.skills.fallback import FallbackSkillV2
from ovos_workshop.skills.ovos import OVOSSkill


class UniversalSkill(OVOSSkill):
    """
    Skill that auto translates input/output from any language

    This skill is designed to automatically translate input and output messages
    between different languages. The intent handlers are ensured to receive
    utterances in the skill's internal language, and they are expected to produce
    utterances in the same internal language.

    The `speak` method will always translate utterances from the internal language
    to the original query language (`self.lang`).

    NOTE: `self.lang` reflects the original query language, but received utterances
          are always in `self.internal_language`.
    """

    def __init__(self, internal_language: str = None,
                 translate_tags: bool = True,
                 autodetect: bool = False,
                 translate_keys: list = None,
                 *args, **kwargs):
        """
        Initialize the UniversalSkill.

        Parameters:
        - internal_language (str): The language in which the skill internally operates.
        - translate_tags (bool): Whether to translate the private __tags__ value (adapt entities).
        - autodetect (bool): If True, the skill will detect the language of the utterance
                            regardless of Session.lang.
        - translate_keys (list): default ["utterance", "utterances"]
                                Keys added here will have values translated in message.data.
        - *args, **kwargs: Additional arguments passed to the parent class constructor.

        Note:
        - If `internal_language` is not provided, it will be set to the language
          specified in the configuration.
        """
        super().__init__(*args, **kwargs)
        # the skill internally only works in this language
        self.internal_language = internal_language
        # __tags__ private value will be translated (adapt entities)
        self.translate_tags = translate_tags
        # keys added here will have values translated in message.data
        self.translate_keys = translate_keys or ["utterance", "utterances"]

        self.autodetect = autodetect
        if not self.internal_language:
            lang = Configuration().get("lang", "en-us")
            LOG.warning(f"UniversalSkill are expected to specify their "
                        f"internal_language, casting to {lang}")
            self.internal_language = lang

    def _load_lang(self, root_directory=None, lang=None):
        """
        unlike base skill class all resources are in self.internal_language by
        default instead of self.lang (which comes from message)
        this ensures things like self.dialog_render reflect self.internal_lang
        """
        lang = lang or self.internal_language  # self.lang in base class
        root_directory = root_directory or self.res_dir
        if lang not in self._lang_resources:
            self._lang_resources[lang] = SkillResources(root_directory, lang,
                                                        skill_id=self.skill_id)
        return self._lang_resources[lang]

    def detect_language(self, utterance: str):
        """
        Detect the language of the given utterance.

        Parameters:
        - utterance (str): The input text whose language needs to be detected.

        Returns:
        str: The detected language code.

        If an error occurs during language detection, it falls back to the language
        specified in `Session.lang`.

        Note:
        - The detected language is based on Session.lang if `self.autodetect` is False.
        - If `self.autodetect` is True, the language is detected regardless of Session.lang.
        - The detection is performed using the configured translator plugin.
        """
        try:
            return self.lang_detector.detect(utterance)
        except Exception as e:
            LOG.error(e)
            # self.lang to account for lang defined in message
            return self.lang.split("-")[0]

    def translate_utterance(self, text: str, target_lang: str, sauce_lang: str = None):
        """
        Translate the given text from the source language to the target language.

        Parameters:
        - text (str): The text to be translated.
        - target_lang (str): The target language code for translation.
        - sauce_lang (str, optional): The source language code. If not provided, it will be detected.

        Returns:
        str: The translated text.

        If the detected source language is the same as the target language, the input
        text is returned unchanged.

        Note:
        - If `self.autodetect` is True, the source language is automatically detected.
        - The translation is performed using the configured translator plugin.
        """
        if self.autodetect:
            sauce_lang = self.detect_language(text)
        else:
            sauce_lang = sauce_lang or self.detect_language(text)
        if sauce_lang.split("-")[0] != target_lang:
            translated = self.translator.translate(text, source=sauce_lang,
                                                   target=target_lang)
            LOG.info("translated " + text + " to " + translated)
            return translated
        return text

    def translate_message(self, message: Message):
        """
        Translate the content of the message to the skill's internal language.

        Parameters:
        - message (Message): The message object to be translated.

        Returns:
        - Message: The translated message.

        The translation process includes translating text, lists, and dictionaries
        based on the configured (per skill) translation keys.
        If enabled, it also translates tags (adapt keywords) in the message data.

        The original and translated data are stored in the `translation_data` attribute
        of the message context.
        """
        # translate speech from input lang to internal lang
        sauce_lang = self.lang  # from message or config
        out_lang = self.internal_language  # skill wants input is in this language,

        if sauce_lang == out_lang and not self.autodetect:
            # do nothing
            return message

        translation_data = {"original": {}, "translated": {},
                            "source_lang": sauce_lang,
                            "internal_lang": self.internal_language}

        def _do_tx(thing):
            if isinstance(thing, str):
                thing = self.translate_utterance(thing, target_lang=out_lang,
                                                 sauce_lang=sauce_lang)
            elif isinstance(thing, list):
                thing = [_do_tx(t) for t in thing]
            elif isinstance(thing, dict):
                thing = {k: _do_tx(v) for k, v in thing.items()}
            return thing

        for key in self.translate_keys:
            if key in message.data:
                translation_data["original"][key] = message.data[key]
                translation_data["translated"][key] = message.data[key] = \
                    _do_tx(message.data[key])

        # special case
        if self.translate_tags:
            translation_data["original"]["__tags__"] = message.data["__tags__"]
            for idx, token in enumerate(message.data["__tags__"]):
                message.data["__tags__"][idx] = \
                    self.translate_utterance(token.get("key", ""),
                                             target_lang=out_lang,
                                             sauce_lang=sauce_lang)
            translation_data["translated"]["__tags__"] = \
                message.data["__tags__"]

        message.context["translation_data"] = translation_data
        return message

    def create_universal_handler(self, handler):
        """
        Create a universal intent handler that translates the message before invoking
        the original handler.

        Parameters:
        - handler (callable): The original intent handler function.

        Returns:
        - callable: A new intent handler function with translation logic.

        The created intent handler takes a message as input, translates its content
        to the skill's internal language using the `translate_message` method, and
        then invokes the original handler with the translated message.

        Manual usage Example:
        ```python
        my_handler = self.create_universal_handler(original_handler_function)
        self.add_event("my_event", my_handler)
        ```

        NOTE: this method should not be used in skill intents,
          that is done automatically for self.register_intent!

          Use only with self.add_event

        """

        def universal_intent_handler(message):
            message = self.translate_message(message)
            LOG.info(get_handler_name(handler))
            handler(message)

        return universal_intent_handler

    def register_intent(self, intent_parser, handler):
        """
        Register an intent with a universal intent handler.

        Parameters:
        - intent_parser (str or IntentParser): The intent parser to register.
        - handler (callable): The original intent handler function.

        This method registers the intent with a universal intent handler, which
        translates the message before invoking the original handler.

        Example:
        ```python
        self.register_intent("my_intent", my_handler_function)
        ```

        See Also:
        - `create_universal_handler` method for creating universal intent handlers.
        """
        handler = self.create_universal_handler(handler)
        super().register_intent(intent_parser, handler)

    def register_intent_file(self, intent_file, handler):
        """
         Register intents from a file with a universal intent handler.

         Parameters:
         - intent_file (str): The path to the intent file.
         - handler (callable): The original intent handler function.

         This method registers intents from a file with a universal intent handler,
         which translates the message before invoking the original handler.

         Example:
         ```python
         self.register_intent_file("my_intents.intent", my_handler_function)
         ```

         See Also:
         - `create_universal_handler` method for creating universal intent handlers.
         """
        handler = self.create_universal_handler(handler)
        super().register_intent_file(intent_file, handler)

    def speak(self, utterance: str, *args, **kwargs):
        """
        Speak the given utterance, translating it if needed.

        Parameters:
        - utterance (str): The text to be spoken.
        - *args, **kwargs: Additional arguments passed to the parent class `speak` method.

        If the output language (`self.lang` / Session.lang) is different from the skill's internal
        language (`self.internal_language`), or autodetection is enabled, the utterance
        is translated before being spoken. The translation data is stored in the meta
        information.

        Example:
        ```python
        self.speak("Hello, how are you?")
        ```

        See Also:
        - `translate_utterance` method for translation logic.
        - `create_universal_handler` method for creating universal intent handlers.
        """
        # translate speech from input lang to output lang
        out_lang = self.lang  # from message or config
        sauce_lang = self.internal_language  # skill output is in this language
        if out_lang != sauce_lang or self.autodetect:
            meta = kwargs.get("meta") or {}
            meta["translation_data"] = {
                "original": utterance,
                "internal_lang": self.internal_language,
                "target_lang": out_lang
            }
            utterance = self.translate_utterance(utterance, sauce_lang, out_lang)
            meta["translation_data"]["translated"] = utterance
            kwargs["meta"] = meta
        super().speak(utterance, *args, **kwargs)

    def _handle_converse_request(self, message: Message):
        message = self.translate_message(message)
        super()._handle_converse_request(message)


class UniversalFallback(UniversalSkill, FallbackSkillV2):
    """
    Fallback Skill that auto translates input/output from any language.

    Fallback handlers are ensured to receive utterances and expected to produce
    responses in `self.internal_language`.

    `self.speak` will always translate utterances from
    `self.internal_lang` to `self.lang`.

    NOTE: `self.lang` reflects the original query language,
          but received utterances are always in `self.internal_language`.
    """

    def create_universal_fallback_handler(self, handler):
        """
        Create a universal fallback handler that translates the message before invoking
        the original fallback handler.

        Parameters:
        - handler (callable): The original fallback handler function.

        Returns:
        - callable: A new fallback handler function with translation logic.

        The created fallback handler takes a message as input, translates its content
        to the skill's internal language using the `translate_message` method, and
        then invokes the original handler with the translated message.
        If the fallback matched the skill is activated with the converse system
        """

        def universal_fallback_handler(message):
            # auto_Translate input
            message = self.translate_message(message)
            LOG.info(get_handler_name(handler))
            success = handler(self, message)
            if success:
                self.activate()
            return success

        return universal_fallback_handler

    def register_fallback(self, handler, priority: int):
        """
        Register a fallback handler with a specified priority.

        Parameters:
        - handler (callable): The original fallback handler function.
        - priority (int): The priority of the fallback handler.

        This method registers the fallback handler with a universal fallback handler,
        which translates the message before invoking the original handler.

        Example:
        ```python
        self.register_fallback(my_fallback_handler_function, priority=5)
        ```

        See Also:
        - `create_universal_fallback_handler` method for creating universal fallback handlers.
        """
        handler = self.create_universal_fallback_handler(handler)
        FallbackSkillV2.register_fallback(self, handler, priority)


class UniversalCommonQuerySkill(UniversalSkill, CommonQuerySkill, ABC):
    """
    CommonQuerySkill that auto translates input/output from any language.

    `CQS_match_query_phrase` and `CQS_action` are ensured to receive phrases in
    `self.internal_language`.

    `CQS_match_query_phrase` is assumed to return a response in `self.internal_language`,
    and it will be translated back before speaking.

    `self.speak` will always translate utterances from
    `self.internal_lang` to `self.lang`.

    NOTE: `self.lang` reflects the original query language,
          but received utterances are always in `self.internal_language`.
    """

    def __handle_query_action(self, message: Message):
        """
        Handle the common query action, translating the message if needed.

        Parameters:
        - message (Message): The message containing the query action.

        This method translates the query phrase to the internal language if the
        output language (`self.lang` / Session.lang) is different or autodetection is enabled.
        Then it invokes the parent method `__handle_query_action`.

        This method is internal and should not be called directly.
        """
        if message.data["skill_id"] != self.skill_id:
            # Not for this skill!
            return
        if self.lang != self.internal_language or self.autodetect:
            message.data["phrase"] = self.translate_utterance(message.data["phrase"],
                                                              sauce_lang=self.lang,
                                                              target_lang=self.internal_language)

        super().__handle_query_action(message)

    def __get_cq(self, search_phrase: str):
        """
         Get a common query result for the given search phrase.

         Parameters:
         - search_phrase (str): The search phrase.

         Returns:
         - tuple or None: A tuple representing the common query result, or None if not found.

         This method converts the input into the internal language if needed, gets
         the common query result, and converts the response back into the source language.

         This method is internal and should not be called directly.
         """
        if self.lang == self.internal_language and not self.autodetect:
            return super().__get_cq(search_phrase)

        # convert input into internal lang
        search_phrase = self.translate_utterance(search_phrase, self.internal_language, self.lang)
        result = super().__get_cq(search_phrase)
        if not result:
            return None
        answer = result[2]
        # convert response back into source lang
        answer = self.translate_utterance(answer, self.lang, self.internal_language)
        if len(result) > 3:
            # optional callback_data
            result = (result[0], result[1], answer, result[3])
        else:
            result = (result[0], result[1], answer)
        return result

    def remove_noise(self, phrase: str, lang: str = None):
        """
        Remove noise to produce the essence of the question.

        Parameters:
        - phrase (str): The input phrase.
        - lang (str, optional): ignored, just for api compat

        Returns:
        - str: The cleaned phrase.

        This method removes noise from the input phrase to extract the essence of the question.
        The method uses the `self.internal_language` as the default language.
        """
        return super().remove_noise(phrase, self.internal_language)



# Copyright 2021 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import abc

from ovos_bus_client.message import Message
from ovos_utils.log import LOG
from ovos_workshop.skills.ovos import OVOSSkill


class IdleDisplaySkill(OVOSSkill):
    """
    equivalent to using @resting_handler decorator in a regular OVOSSkill

    Helper class for skills that define an idle display.

    An idle display is what shows on a device's screen when it is not in use
    by other skills. i.e. a Home Screen skill.

    The idle display should show when no other skill is using the display. Some
    skills use the display for a defined period of time before returning to the
    idle display (e.g. Weather Skill). Some skills take control of the display
    indefinitely (e.g. Timer Skill).

    The display could be a touch screen (such as on the Mark II), or an
    Arduino LED array (such as on the Mark I), or any other type of display.
    This base class is meant to be agnostic to the type of display.
    """

    @abc.abstractmethod
    def handle_idle(self):
        """
        Override this method to display the idle screen.
        """
        raise NotImplementedError("Subclass must override the handle_idle method")

    def _register_system_event_handlers(self):
        """
        Defines the bus events handled in this skill and their handlers.
        """
        super()._register_system_event_handlers()
        self.add_event("homescreen.manager.activate.display", self.handle_homescreen_request)
        self.add_event("homescreen.manager.reload.list", self.register_homescreen)
        self.add_event("mycroft.skills.shutdown", self._remove_homescreen_on_shutdown)
        self.register_homescreen()

    def register_homescreen(self, message: Message = None):
        """
        Update the internal _homescreen_entry object
        for this skill and send it to the Home Screen Manager.
        @param message: optional Message associated with request
        """
        LOG.debug(f"Registering Homescreen {self.skill_id}")
        self.bus.emit(Message("homescreen.manager.add",
                              {"class": "IdleDisplaySkill",  # TODO - rm in ovos-gui, only for compat
                               "id": self.skill_id}))

    def remove_homescreen(self, message: Message):
        """
        Remove this skill's homescreen_entry from the Home Screen Manager
        @param message: `mycroft.skills.shutdown` message
        """
        LOG.debug(f"Requesting homescreen removal of {self.skill_id}")
        msg = message.forward("homescreen.manager.remove",
                              {"id": self.skill_id})
        self.bus.emit(msg)

    def _remove_homescreen_on_shutdown(self, message: Message):
        """
        Remove this homescreen from the Home Screen Manager if requested
        @param message: `mycroft.skills.shutdown` message
        """
        if message.data["id"] == self.skill_id:
            self.remove_homescreen(message)

    def handle_homescreen_request(self, message: Message):
        """
        Display this home screen if requested by the Home Screen Manager
        @param message: `homescreen.manager.activate.display` message
        """
        if message.data["homescreen_id"] == self.skill_id:
            self.handle_idle()
            self.bus.emit(message.reply("skill.idle.displayed"))



# Copyright 2019 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import shutil
from os.path import join, exists

from ovos_bus_client import MessageBusClient, Message
from ovos_utils.log import LOG, log_deprecation, deprecated
from ovos_workshop.decorators.compat import backwards_compat
from ovos_workshop.skills.ovos import OVOSSkill, is_classic_core, _OVOSSkillMetaclass


class _SkillMetaclass(_OVOSSkillMetaclass):
    """
    This metaclass ensures we can load skills like regular python objects.
    mycroft-core required a skill loader helper class, which created the skill
    and then finished object init. This meant skill_id and bus were not
    available in init method, so mycroft introduced a method named `initialize`
    that was called after `skill_id` and `bus` were defined.

    To make skills pythonic and standalone, this metaclass is used to auto init
    old skills and help in migrating to new standards.

    To override isinstance checks we also need to use a metaclass

    TODO: remove compat ovos-core 0.2.0 at the latest, including MycroftSkill class
    """

    def __instancecheck_classic__(self, instance):
        # instance imported from vanilla mycroft
        from mycroft.skills import MycroftSkill as _CoreSkill
        from ovos_workshop.app import OVOSAbstractApplication
        if issubclass(instance.__class__, _CoreSkill):
            return True
        return issubclass(instance.__class__, OVOSSkill) and \
            not issubclass(instance.__class__, OVOSAbstractApplication)

    @backwards_compat(classic_core=__instancecheck_classic__)
    def __instancecheck__(self, instance):
        from ovos_workshop.app import OVOSAbstractApplication
        if issubclass(instance.__class__, OVOSAbstractApplication):
            return False
        return super().__instancecheck__(instance) or \
            issubclass(instance.__class__, OVOSSkill)

    def __call__(cls, *args, **kwargs):
        from ovos_bus_client import MessageBusClient
        from ovos_utils.messagebus import FakeBus
        bus = None
        skill_id = None

        if "bus" not in kwargs:
            for a in args:
                if isinstance(a, MessageBusClient) or isinstance(a, FakeBus):
                    bus = a
                    LOG.warning(
                        f"bus should be a kwarg, guessing {a} is the bus")
                    break
            else:
                LOG.warning("skill initialized without bus!! this is legacy "
                            "behaviour and requires you to call skill.bind(bus)"
                            " or skill._startup(skill_id, bus)\n"
                            "bus will be required starting on ovos-core 0.1.0")
                return super().__call__(*args, **kwargs)

        if "skill_id" in kwargs:
            skill_id = kwargs.pop("skill_id")
        if "bus" in kwargs:
            bus = kwargs.pop("bus")
        if not skill_id:
            LOG.warning(f"skill_id should be a kwarg, please update "
                        f"{cls.__name__}")
            if args and isinstance(args[0], str):
                a = args[0]
                if a[0].isupper():
                    # in mycroft name is CamelCase by convention, not skill_id
                    LOG.debug(f"ambiguous skill_id, ignoring {a} as it appears "
                              f"to be a CamelCase name")
                else:
                    LOG.warning(f"ambiguous skill_id, assuming positional "
                                f"argument: {a}")
                    skill_id = a

            if not skill_id:
                LOG.warning("skill initialized without bus!! this is legacy "
                            "behaviour and requires you to call skill.bind(bus)"
                            " or skill._startup(skill_id, bus)\n"
                            "bus will be required starting on ovos-core 0.1.0")
                return super().__call__(*args, **kwargs)

                # by convention skill_id is the folder name
                # usually repo.author
                # TODO - uncomment once above is deprecated
                # skill_id = dirname(inspect.getfile(cls)).split("/")[-1]
                # LOG.warning(f"missing skill_id, assuming folder name "
                #             f"convention: {skill_id}")

        try:
            # skill follows latest best practices,
            # accepts kwargs and does its own init
            return super().__call__(skill_id=skill_id, bus=bus, **kwargs)
        except TypeError:
            LOG.warning(f"Legacy skill signature detected for {skill_id};"
                        f" attempting to init skill manually, self.bus and "
                        f"self.skill_id will only be available in "
                        f"self.initialize. `__init__` method needs to accept "
                        f"`skill_id` and `bus` to resolve this.")

        # skill did not update its init method, init it manually
        # NOTE: no try/except because all skills must accept this initialization
        # this is what skill loader does internally
        skill = super().__call__(*args, **kwargs)
        skill._startup(bus, skill_id)
        return skill


class MycroftSkill(OVOSSkill, metaclass=_SkillMetaclass):
    """
    Base class for mycroft skills providing common behaviour and parameters
    to all Skill implementations. This class is kept for backwards-compat. It is
    recommended to implement `OVOSSkill` to properly implement new methods.
    """

    @deprecated("mycroft-core has been deprecated, please move to ovos-core", "0.1.0")
    def __classic_init__(self, name: str = None, bus: MessageBusClient = None,
                         use_settings: bool = True, *args, **kwargs):
        """
        Create a MycroftSkill object.
        @param name: DEPRECATED skill_name
        @param bus: MessageBusClient to bind to skill
        @param use_settings: DEPRECATED option to disable settings sync
        """
        super().__init__(name=name, bus=bus, *args, **kwargs)

        self._initial_settings = {}
        self.settings_write_path = None
        self.settings_manager = None

        # old kludge from fallback skills, unused according to grep
        if use_settings is False:
            log_deprecation("use_settings has been deprecated! "
                            "skill settings are always enabled", "0.1.0")

        self.settings_write_path = self.root_dir

    @backwards_compat(classic_core=__classic_init__)
    @deprecated("MycroftSkill class has been deprecated, please subclass from OVOSSkill", "0.1.0")
    def __init__(self, name: str = None, bus: MessageBusClient = None,
                 use_settings: bool = True, *args, **kwargs):
        """
        Create a MycroftSkill object.
        @param name: DEPRECATED skill_name
        @param bus: MessageBusClient to bind to skill
        @param use_settings: DEPRECATED option to disable settings sync
        """
        super().__init__(name=name, bus=bus, *args, **kwargs)
        self._initial_settings = {}
        self.settings_write_path = None
        self.settings_manager = None

    def __init_settings_manager_classic(self):
        super()._init_settings_manager()
        from mycroft.skills.settings import SettingsMetaUploader
        self._settings_meta = SettingsMetaUploader(self.root_dir,
                                                   self.skill_id)

    def __init_settings_manager_standalone(self):
        super()._init_settings_manager()

    @backwards_compat(classic_core=__init_settings_manager_classic,
                      no_core=__init_settings_manager_standalone)
    def _init_settings_manager(self):
        super()._init_settings_manager()
        # backwards compat - self.settings_meta has been deprecated
        # in favor of settings manager
        from mycroft.deprecated.skills.settings import SettingsMetaUploader
        self._settings_meta = SettingsMetaUploader(self.root_dir,
                                                   self.skill_id)

    def __init_settings_classic(self):
        # migrate settings if needed
        if not exists(self.settings_path) and \
                exists(self._old_settings_path):
            LOG.warning("Found skill settings at pre-xdg location, "
                        "migrating!")
            shutil.copy(self._old_settings_path, self.settings_path)
            LOG.info(f"{self._old_settings_path} moved to "
                     f"{self.settings_path}")
        super()._init_settings()

    @backwards_compat(classic_core=__init_settings_classic)
    def _init_settings(self):
        """Setup skill settings."""
        super()._init_settings()

    # patched due to functional (internal) differences under mycroft-core
    def __on_end_classic(self, message: Message, handler_info: str,
                         skill_data: dict):
        # mycroft-core style settings
        if self.settings != self._initial_settings:
            try:
                from mycroft.skills.settings import save_settings
                save_settings(self.settings_write_path, self.settings)
                self._initial_settings = dict(self.settings)
            except Exception as e:
                LOG.exception(f"Failed to save skill settings: {e}")
        if handler_info:
            msg_type = handler_info + '.complete'
            message.context["skill_id"] = self.skill_id
            self.bus.emit(message.forward(msg_type, skill_data))

    @backwards_compat(classic_core=__on_end_classic)
    def _on_event_end(self, message: Message, handler_info: str,
                      skill_data: dict):
        """
        Store settings and indicate that the skill handler has completed
        """
        return super()._on_event_end(message, handler_info, skill_data)

    # refactored - backwards compat + log warnings
    @property
    def settings_meta(self):
        log_deprecation("Use `self.settings_manager`", "0.1.0")
        return self._settings_meta

    # refactored - backwards compat + log warnings
    @settings_meta.setter
    def settings_meta(self, val):
        log_deprecation("Use `self.settings_manager`", "0.1.0")
        self._settings_meta = val

    # internal - deprecated under ovos-core
    @property
    def _old_settings_path(self):
        log_deprecation("This path is no longer used", "0.1.0")
        old_dir = self.config_core.get("data_dir") or "/opt/mycroft"
        old_folder = self.config_core.get("skills", {}).get("msm", {}) \
                         .get("directory") or "skills"
        return join(old_dir, old_folder, self.skill_id, 'settings.json')

    # patched due to functional (internal) differences under mycroft-core
    def __get_settings_pclassic(self):
        if self.settings_write_path and \
                self.settings_write_path != self.root_dir:
            log_deprecation("`self.settings_write_path` is no longer used",
                            "0.1.0")
            return join(self.settings_write_path, 'settings.json')
        return super().settings_path

    @property
    @backwards_compat(classic_core=__get_settings_pclassic)
    def settings_path(self):
        return super().settings_path
